use std::sync::Arc;

use cpal::traits::{DeviceTrait, HostTrait, StreamTrait};
use protocol::Error;
use tokio::sync::{broadcast, mpsc};

use crate::gui::State;

pub async fn record(
    mut record_control: mpsc::Receiver<State>,
    record_sender: Arc<broadcast::Sender<f32>>,
) -> Result<(), Error> {
    let host = cpal::default_host();
    let input_device = host
        .default_input_device()
        .ok_or(Error::Record("Input Device".to_string()))?;
    let config = input_device
        .default_input_config()
        .map_err(|inner| Error::Record(inner.to_string()))?
        .into();
    println!("Recorder Stream Config = {:#?}", config);

    let input = move |data: &[f32], _: &cpal::InputCallbackInfo| {
        for &sample in data {
            if record_sender.receiver_count() > 0 && sample != 0.0 {
                if let Err(err_val) = record_sender.send(sample) {
                    eprintln!("Error: Microphone Send | {}", err_val);
                }
            }
        }
    };

    let input_stream = input_device
        .build_input_stream(&config, input, voice_error, None)
        .map_err(|inner| Error::Record(inner.to_string()))?;

    input_stream
        .pause()
        .map_err(|inner| Error::Record(inner.to_string()))?;

    tokio::task::block_in_place(|| {
        loop {
            match record_control.blocking_recv() {
                Some(message) => match message {
                    State::Active => input_stream
                        .play()
                        .map_err(|inner| Error::Record(inner.to_string()))?,
                    State::Passive => input_stream
                        .pause()
                        .map_err(|inner| Error::Record(inner.to_string()))?,
                    State::Loading => {}
                },
                None => {
                    input_stream
                        .pause()
                        .map_err(|inner| Error::Record(inner.to_string()))?;
                    return Ok(());
                }
            }
        }
    })
}

pub async fn play(
    mut play_control: mpsc::Receiver<State>,
    mut play_receiver: broadcast::Receiver<f32>,
) -> Result<(), Error> {
    let host = cpal::default_host();
    let output_device = host.default_output_device().unwrap();
    let config = output_device.default_output_config().unwrap().into();
    println!("Speaker Stream Config = {:#?}", config);

    let output = move |data: &mut [f32], _: &cpal::OutputCallbackInfo| {
        for sample in data {
            if play_receiver.len() > 0 {
                match play_receiver.blocking_recv() {
                    Ok(received_sample) => *sample = received_sample,
                    Err(err_val) => match err_val {
                        broadcast::error::RecvError::Closed => {
                            eprintln!("Error: Speaker Receive | Local Channel | Channel Closed");
                            return;
                        }
                        broadcast::error::RecvError::Lagged(lag_amount) => {
                            eprintln!(
                                "Error: Speaker Receive | Local Channel | Lagging by -> {}",
                                lag_amount
                            );
                            play_receiver = play_receiver.resubscribe();
                        }
                    },
                }
            } else {
                *sample = 0.0
            }
        }
    };

    let output_stream = output_device
        .build_output_stream(&config, output, voice_error, None)
        .unwrap();

    output_stream
        .play()
        .map_err(|inner| Error::Play(inner.to_string()))?;

    tokio::task::block_in_place(|| {
        loop {
            match play_control.blocking_recv() {
                Some(message) => match message {
                    State::Active => output_stream
                        .play()
                        .map_err(|inner| Error::Play(inner.to_string()))?,
                    State::Passive => output_stream
                        .pause()
                        .map_err(|inner| Error::Play(inner.to_string()))?,
                    State::Loading => {}
                },
                None => {
                    output_stream
                        .pause()
                        .map_err(|inner| Error::Play(inner.to_string()))?;
                    return Ok(());
                }
            }
        }
    })
}

fn voice_error(err_val: cpal::StreamError) {
    eprintln!("Error: Voice Error | {}", err_val);
}
