use std::{
    collections::VecDeque,
    path::Path,
    sync::{Arc, LazyLock},
    time::Duration,
};

use s2n_quic::{
    Connection, Server,
    stream::{ReceiveStream, SendStream},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    sync::{RwLock, broadcast},
    time::sleep,
};

use crate::ServerConfig;
const BUFFER_LENGTH: usize = 1024 * 16 * 2;

#[derive(Debug)]
struct User {
    audio_buffer: Arc<RwLock<VecDeque<f32>>>,
}

impl User {
    async fn new() -> Arc<RwLock<VecDeque<f32>>> {
        let audio_buffer = Arc::new(RwLock::new(VecDeque::new()));
        let new_user = Self {
            audio_buffer: audio_buffer.clone(),
        };
        ONLINE_USERS.write().await.push(new_user);
        audio_buffer
    }
}
static ONLINE_USERS: LazyLock<RwLock<Vec<User>>> = LazyLock::new(|| vec![].into());
static GLOBAL_MIXER: LazyLock<broadcast::Sender<f32>> =
    LazyLock::new(|| broadcast::channel(BUFFER_LENGTH).0);

async fn global_mixer() {
    let global_mixer_sender = GLOBAL_MIXER.clone();

    loop {
        sleep(Duration::from_millis(100)).await;
        let mut mixed_audio_buffer = VecDeque::new();
        for online_user in ONLINE_USERS.read().await.iter() {
            let mut inner_buffer = vec![];
            while let Some(audio_data) = online_user.audio_buffer.write().await.pop_front() {
                inner_buffer.push(audio_data);
            }

            for (i, audio_data) in inner_buffer.iter().enumerate() {
                match mixed_audio_buffer.get(i) {
                    Some(original_value) => mixed_audio_buffer[i] = original_value + audio_data,
                    None => mixed_audio_buffer.push_back(*audio_data),
                }
            }
        }

        mixed_audio_buffer.truncate(mixed_audio_buffer.len() * 85 / 100);
        while let Some(audio_data) = mixed_audio_buffer.pop_front() {
            if let Err(err_val) = global_mixer_sender.send(audio_data) {
                eprintln!("Error: Send Mixed Audio Data | Local | {}", err_val);
            }
        }
    }
}

pub async fn start(server_config: ServerConfig) {
    let mut server = Server::builder()
        .with_io(server_config.server_address.as_str())
        .unwrap()
        .with_tls((
            Path::new(&server_config.certificate_path),
            Path::new(&server_config.certificate_key_path),
        ))
        .unwrap()
        .start()
        .unwrap();

    tokio::spawn(global_mixer());

    while let Some(connection) = server.accept().await {
        tokio::spawn(handle_client(connection));
    }
}

async fn handle_client(mut connection: Connection) {
    println!(
        "Server Name = {}",
        connection.server_name().unwrap().unwrap().to_string()
    );

    println!(
        "Client Address = {}",
        connection.remote_addr().unwrap().to_string()
    );

    let stream = connection
        .accept_bidirectional_stream()
        .await
        .unwrap()
        .unwrap();
    let (receive_stream, send_stream) = stream.split();

    let user_audio_buffer = User::new().await;

    tokio::spawn(receiver_audio_data(receive_stream, user_audio_buffer));
    tokio::spawn(send_audio_data(send_stream));
}

async fn send_audio_data(mut send_stream: SendStream) {
    let mut global_mixer_receiver = GLOBAL_MIXER.subscribe();

    loop {
        match global_mixer_receiver.recv().await {
            Ok(received_audio_data) => {
                if let Err(err_val) = send_stream.write_f32(received_audio_data).await {
                    eprintln!("Error: Send Audio Data | Remote | {}", err_val);
                    return;
                }
            }
            Err(err_val) => {
                eprintln!("Error: Receive Mixed Audio Data | Local | {}", err_val);
                return;
            }
        }
    }
}

async fn receiver_audio_data(
    mut receive_stream: ReceiveStream,
    audio_buffer: Arc<RwLock<VecDeque<f32>>>,
) {
    loop {
        match receive_stream.read_f32().await {
            Ok(received_audio_data) => {
                audio_buffer.write().await.push_back(received_audio_data);
            }
            Err(err_val) => {
                eprintln!("Error: Receive Audio Data | Remote | {}", err_val);
                return;
            }
        }
    }
}
